from sqlalchemy import create_engine, text
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import QueuePool

from contextlib import contextmanager

from .models import Base


class Database:
    def __init__(self, url: str):
        # Create engine with connection pooling
        self.engine = create_engine(
            url,
            poolclass=QueuePool,
            pool_size=5,  # Maximum number of connections in the pool
            max_overflow=10,  # Maximum number of connections that can be created beyond pool_size
            pool_timeout=30,  # Timeout for getting a connection from the pool
            pool_pre_ping=True,  # Enable connection health checks
        )

        # Create session factory
        self.SessionLocal = sessionmaker(
            autocommit=False, autoflush=False, bind=self.engine
        )

    @contextmanager
    def get_session(self):
        """
        Get a database session. Use with a context manager:

        with db.get_session() as session:
            session.query(...)
        """
        session = self.SessionLocal()
        try:
            yield session
        finally:
            session.close()

    def create_tables(self):
        """Create all tables defined in models"""
        Base.metadata.create_all(bind=self.engine)

    def check_connection(self):
        """Check if the connection to the database is working"""
        try:
            with self.get_session() as session:
                session.execute(text("SELECT 1"))
        except Exception as e:
            raise e



import datetime as dt
import enum
import random
import string
from abc import abstractmethod

from sqlalchemy import MetaData, Column, text
from sqlalchemy.inspection import inspect
from sqlalchemy.orm import DeclarativeBase, declared_attr, relationship
from sqlalchemy.dialects.postgresql import TIMESTAMP, TEXT

convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "%(table_name)s_%(column_0_name)s_unique",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "%(table_name)s_%(column_0_name)s_fkey",
    "pk": "%(table_name)s_pkey",
}

metadata = MetaData(naming_convention=convention)


class Base(DeclarativeBase):
    metadata = metadata

    @classmethod
    def __declare_last__(cls):
        for mapper in cls.registry.mappers:
            for rel in mapper.relationships:
                if "passive_deletes" not in rel.info:
                    rel.passive_deletes = True
                if "cascade" not in rel.info:
                    # For one-to-many relationships, use delete-orphan
                    if rel.direction.name == "ONETOMANY":
                        rel.cascade = "all, delete-orphan"
                    # For many-to-one or many-to-many, use just 'all'
                    else:
                        rel.cascade = "all"

    def __repr__(self):
        """
        This method is used to represent the model instance in a more
        readable format.
        """
        try:
            identity = inspect(self).identity
        except:
            identity = self.__dict__
        return f"<{self.__class__.__name__} {identity}>"


class HumanIDMixin:
    __prefix__: str
    __id_length__: int = 16

    @classmethod
    def gen_obj_id(cls, prefix: str = None, length: int = 16):
        """Generate a human-readable ID with prefix"""
        if prefix is None:
            prefix = cls.__prefix__

        characters = string.ascii_letters + string.digits
        random_string = "".join(random.choices(characters, k=length))
        return f"{prefix}_{random_string}"

    @classmethod
    def generate_id(cls):
        return cls.gen_obj_id()



from setuptools import setup, find_packages

setup(
    name="solai-db",
    packages=find_packages(
        include=["solai_db", "solai_db.*"]
    ),  # This will include all subpackages
    version="0.1.11",
    author="SolAI",
    description="The database for SolAI",
    install_requires=[
        "sqlalchemy>=2.0.0",
        "psycopg2-binary>=2.9.0",
    ],
)



from .base import Base, HumanIDMixin
from sqlalchemy import Column, ForeignKey, text
from sqlalchemy.orm import relationship
from sqlalchemy.dialects.postgresql import (
    TEXT,
    JSONB,
    INTEGER,
    BOOLEAN,
    ARRAY,
    TIMESTAMP,
)


class Integration(Base):
    __tablename__ = "integrations"

    id = Column(TEXT, primary_key=True)
    name = Column(TEXT, nullable=False)

    # Relationships
    inverters = relationship("Inverter", back_populates="integration")
    tokens = relationship("Token", back_populates="integration")


class Inverter(Base, HumanIDMixin):
    __tablename__ = "inverters"
    __prefix__ = "inv"
    __id_length__ = 16

    id = Column(TEXT, primary_key=True)
    created_at = Column(
        TIMESTAMP(timezone=True),
        server_default=text("now()"),
        nullable=False,
    )
    name = Column(TEXT, nullable=False)
    external_id = Column(TEXT, nullable=False)
    rated_power = Column(INTEGER, nullable=False)
    data = Column(JSONB, nullable=False)

    token_id = Column(
        TEXT,
        ForeignKey("token_vault.tokens.id", ondelete="CASCADE", onupdate="CASCADE"),
        nullable=False,
    )

    # Foreign Keys
    integration_id = Column(
        TEXT,
        ForeignKey("integrations.id", ondelete="CASCADE", onupdate="CASCADE"),
        nullable=False,
    )

    # Relationships
    integration = relationship("Integration", back_populates="inverters")
    token = relationship("Token", back_populates="inverters")
    measurements = relationship("InverterMeasurement", back_populates="inverter")


class InverterMeasurement(Base):
    __tablename__ = "inverter_measurements"

    inverter_id = Column(
        TEXT,
        ForeignKey("inverters.id", ondelete="CASCADE", onupdate="CASCADE"),
        primary_key=True,
    )
    time = Column(TIMESTAMP(timezone=True), primary_key=True)
    power = Column(INTEGER, nullable=False)

    inverter = relationship("Inverter", back_populates="measurements")



from .base import Base, HumanIDMixin
from sqlalchemy import Column, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.dialects.postgresql import TEXT, INTEGER, TIMESTAMP
from sqlalchemy.sql import text
import datetime as dt


class Token(Base, HumanIDMixin):
    __tablename__ = "tokens"
    __table_args__ = {"schema": "token_vault"}
    __prefix__ = "tkn"
    __id_length__ = 16

    id = Column(TEXT, primary_key=True)
    created_at = Column(
        TIMESTAMP(timezone=True),
        server_default=text("now()"),
        nullable=False,
    )
    access_token = Column(TEXT, nullable=False)
    refresh_token = Column(TEXT, nullable=False)
    expires_in = Column(INTEGER, nullable=False)
    token_type = Column(TEXT, nullable=False)
    scope = Column(TEXT, nullable=False)
    updated_at = Column(TIMESTAMP(timezone=True), nullable=False)

    # Foreign Keys
    integration_id = Column(
        TEXT, ForeignKey("integrations.id", ondelete="CASCADE", onupdate="CASCADE")
    )
    # Relationships
    integration = relationship("Integration", back_populates="tokens")
    inverters = relationship("Inverter", back_populates="token")

    def is_expired(self):
        return self.updated_at + dt.timedelta(
            seconds=self.expires_in
        ) < dt.datetime.now(tz=dt.timezone.utc)
